from airflow import DAG
from airflow.operators.python import PythonOperator
from datetime import datetime, timedelta
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import VotingClassifier
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score
from sklearn.impute import SimpleImputer
import mlflow
import mlflow.sklearn
import os

# Compute Population Stability Index (PSI) for numeric features
def compute_psi(original, drifted, bins=10):
    # Convert to numeric, coercing errors to NaN and then drop NaN values
    original = pd.to_numeric(original, errors='coerce').dropna()
    drifted = pd.to_numeric(drifted, errors='coerce').dropna()

    # Ensure there are no NaN values after conversion
    if original.isnull().any() or drifted.isnull().any():
        raise ValueError("Data contains NaN values after conversion")

    original_counts, _ = np.histogram(original, bins=bins)
    drifted_counts, _ = np.histogram(drifted, bins=bins)
    
    # Avoid division by zero or log of zero by adding a small epsilon
    epsilon = 1e-8
    original_percents = original_counts / sum(original_counts) + epsilon
    drifted_percents = drifted_counts / sum(drifted_counts) + epsilon
    
    psi_values = (original_percents - drifted_percents) * np.log(original_percents / drifted_percents)
    return sum(psi_values)

def simulate_data_drift():
    """
    Simulate data drift by perturbing features.
    """
    # Load original data
    data_path = os.getenv("DATA_PATH", "/opt/airflow/data/processed_data")
    csv_files = [os.path.join(data_path, file) for file in os.listdir(data_path) if file.endswith(".csv")]
    data = pd.concat([pd.read_csv(file) for file in csv_files], ignore_index=True)

    # Add drift by modifying distributions
    drifted_data = data.copy()
    drifted_data['TotalCharges'] = drifted_data['TotalCharges'] * np.random.uniform(0.9, 1.1, len(drifted_data))
    drifted_data['MonthlyCharges'] = drifted_data['MonthlyCharges'] * np.random.uniform(0.8, 1.2, len(drifted_data))

    # Save drifted data
    drifted_data.to_csv("/opt/airflow/data/drifted_churn.csv", index=False)
    print("Data drift simulation complete.")

def detect_data_drift():
    """
    Detect data drift using statistical metrics.
    """
    original_data = pd.read_csv("/opt/airflow/data/churn.csv")
    drifted_data = pd.read_csv("/opt/airflow/data/drifted_churn.csv")

    psi_total_charges = compute_psi(original_data['TotalCharges'], drifted_data['TotalCharges'])
    psi_monthly_charges = compute_psi(original_data['MonthlyCharges'], drifted_data['MonthlyCharges'])

    # Threshold for drift detection
    drift_threshold = 0.1
    drift_detected = (psi_total_charges > drift_threshold) or (psi_monthly_charges > drift_threshold)
    print(f"PSI TotalCharges: {psi_total_charges}, PSI MonthlyCharges: {psi_monthly_charges}")
    return drift_detected

def retrain_model():
    """
    Retrain the model using drifted data and log the updated model to MLflow.
    """
    # Load drifted data
    data = pd.read_csv("/opt/airflow/data/drifted_churn.csv")

    data = data.drop(columns=['customerID'])

    # Split into features and target
    X = data.drop(columns=['Churn'])
    y = data['Churn'].map({'Yes': 1, 'No': 0})

    # Handle missing values by imputing them with the mean strategy
    imputer = SimpleImputer(strategy='mean')
    X = imputer.fit_transform(X)  # Fit and transform the training data

    # Split into train and test sets
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # Define individual models
    logistic_model = LogisticRegression(max_iter=1000, random_state=42)
    random_forest_model = RandomForestClassifier(n_estimators=100, random_state=42)
    gradient_boosting_model = GradientBoostingClassifier(n_estimators=100, random_state=42)

    ensemble_model = VotingClassifier(
        estimators=[('logistic', logistic_model),
                    ('random_forest', random_forest_model),
                    ('gradient_boosting', gradient_boosting_model)],
        voting='soft'
    )

    # Train the model
    ensemble_model.fit(X_train, y_train)
    y_pred = ensemble_model.predict(X_test)

    # Evaluate model
    accuracy = accuracy_score(y_test, y_pred)
    print(f"Retrained Model Accuracy: {accuracy}")

    # Log updated model to MLflow
    mlflow.set_tracking_uri("http://mlflow:5000")
    with mlflow.start_run():
        mlflow.log_metric("accuracy", accuracy)
        mlflow.sklearn.log_model(ensemble_model, "retrained_model")
        print("Model retraining and logging complete.")

default_args = {
    'owner': 'airflow',
    'depends_on_past': False,
    'email_on_failure': False,
    'email_on_retry': False,
    'retries': 1,
    'retry_delay': timedelta(minutes=5),
}

with DAG(
    'data_drift_detection_and_retraining',
    default_args=default_args,
    description='Detect data drift and retrain model when threshold is reached',
    schedule_interval=None,
    start_date=datetime(2025, 1, 1),
    catchup=False,
) as dag:

    simulate_drift_task = PythonOperator(
        task_id='simulate_data_drift',
        python_callable=simulate_data_drift,
    )

    detect_drift_task = PythonOperator(
        task_id='detect_data_drift',
        python_callable=detect_data_drift,
    )

    retrain_task = PythonOperator(
        task_id='retrain_model',
        python_callable=retrain_model,
    )

    simulate_drift_task >> detect_drift_task >> retrain_task
