from pyspark.sql import SparkSession
from pyspark.sql.functions import col, when
from functools import reduce
from pyspark.ml.feature import StringIndexer
import os

spark = SparkSession.builder \
    .appName("Telco Customer Churn Analysis") \
    .getOrCreate()

file_path = "/opt/airflow/data/churn.csv"
telco_df = spark.read.csv(file_path, header=True, inferSchema=True)

# Feature Engineering
# Create tenure buckets
telco_df = telco_df.withColumn(
    "tenure_bucket",
    when(col("tenure") <= 12, "short-term")
    .when((col("tenure") > 12) & (col("tenure") <= 36), "mid-term")
    .otherwise("long-term")
)

# Calculate Average Monthly Spend
telco_df = telco_df.withColumn(
    "avg_monthly_spend",
    when(col("tenure") > 0, col("TotalCharges") / col("tenure")).otherwise(0)
)

# Count the number of 'Yes' in service-related columns
service_columns = ["PhoneService", "MultipleLines", "OnlineSecurity", "OnlineBackup", 
                   "DeviceProtection", "TechSupport", "StreamingTV", "StreamingMovies"]

service_count_expr = reduce(
    lambda acc, c: acc + when(col(c) == "Yes", 1).otherwise(0),
    service_columns,
    when(col(service_columns[0]) == "Yes", 1).otherwise(0) * 0  # Start with 0
)

telco_df = telco_df.withColumn("service_count", service_count_expr)

# Calculate Customer Lifetime Value (CLV)
telco_df = telco_df.withColumn(
    "CLV",
    col("MonthlyCharges") * col("tenure")
)

# Loyalty flag
telco_df = telco_df.withColumn(
    "is_loyal",
    when(col("tenure") > 36, 1).otherwise(0)
)

# Encode categorical columns
categorical_columns = ["gender", "Partner", "Dependents", "tenure", "SeniorCitizen", "InternetService",
                       "Contract", "PaperlessBilling", "PhoneService", "MultipleLines", "OnlineSecurity", "OnlineBackup", 
                       "PaymentMethod", "DeviceProtection", "TechSupport", "StreamingTV", "StreamingMovies", "tenure_bucket"]

for col_name in categorical_columns:
    indexer = StringIndexer(inputCol=col_name, outputCol=col_name + "_index")
    telco_df = indexer.fit(telco_df).transform(telco_df)

telco_df = telco_df.drop(*categorical_columns)

telco_df = telco_df.dropna()

output_path = "/opt/airflow/data/processed_data"
if not os.path.exists(output_path):
    os.makedirs(output_path)

telco_df.write.csv(output_path, header=True, mode="overwrite")
print(f"Feature engineered and encoded data saved to {output_path}")
